#ifndef MPIPE_API_DEVICE_H
#define MPIPE_API_DEVICE_H

#include "mpipe_types.h"
#include "mpipe_stream.h"

EXTERN_C_BEGIN

struct mpipe_command_t;
struct mpipe_media_format_t;

// параметры stun сервера
typedef struct mpipe_ice_server_t
{
    const char*                         url;
    const char*                         user;
    const char*                         password;
} mpipe_ice_server_t;

// параметры битрейта rtc-трека
typedef struct mpipe_track_bitrate_params_t
{
    uint32_t                            min_bitrate;    // минимальный битрейт в bps (0 - оставить значение по умолчанию)
    uint32_t                            max_bitrate;    // максимальный битрейт в bps (0 - оставить значение по умолчанию)

} mpipe_track_bitrate_params_t;

// параметры трека для rtc-device
typedef struct mpipe_track_params_t
{
    enum mpipe_media_type_t             media_type;     // тип медиа
    enum mpipe_track_direction_t        direction;      // направление
    const char*                         track_id;       // идентификатор (любой)
    mpipe_track_bitrate_params_t        bitrate;        // параметры битрейта для трека (только для отправителя)

} mpipe_rtc_track_params_t;


// описатель медиа-устройства
typedef struct mpipe_media_device_t
{
    mpipe_handle_t                      handle;         // номер дескриптора
    enum mpipe_device_type_t            device_type;    // тип устройства: input, output, webrtc
    enum mpipe_device_state_t           state;          // состояние устройства
    void*                               user_data;      // пользовательский контекст

} mpipe_media_device_t;

// параметры для устройства ввода (камеры, файлы, микрофоны и т.д.)
typedef struct mpipe_device_params_input_t
{
    const char*                         url;            // url устройства
    const char*                         options;        // специальные опции (зависят от типа)

} mpipe_device_params_input_t;

// параметры для устройства вывода (файлы, динамики, rtmp-публикации)
typedef struct mpipe_device_params_output_t
{
    const char*                         url;            // url
    const char*                         options;        // специальные опции (зависят от типа)

    const mpipe_media_format_t**        streams;        // NULL-termination массив выходных форматов (для выходных устрйоств надо задавать форматы вывода)

} mpipe_device_params_output_t;

// параметры битрейта для p2p-соединения
typedef struct mpipe_peer_bitrate_params_t
{
    uint32_t                            start_bitrate;  // начальный битрейт в bps (0 - оставить значение по умолчанию)
    uint32_t                            min_bitrate;    // минимальный битрейт в bps (0 - оставить значение по умолчанию)
    uint32_t                            max_bitrate;    // максимальный битрейт в bps (0 - оставить значение по умолчанию)

} mpipe_peer_bitrate_params_t;

// параметры для устройоств типа webrtc
typedef struct mpipe_device_params_rtc_t
{
    const mpipe_ice_server_t**          ice_servers;    // NULL-termination массив публичных stun-серверов (если необходима работа за NAT)
    uint16_t                            min_rtp_port;   // нижний порог диапазона RTP-портов
    uint16_t                            max_rtp_port;   // верхний порог диапазона RTP-портов
    const char*                         data_channel;   // идентификатор канала данных (если NULL - канал данных неиспользуется)
    const char*                         nat_table;      // таблица ручной трансляции адресов в формате: AddressFrom1:AddressTo1;AddressFrom2:AddressTo2
    mpipe_peer_bitrate_params_t         bitrate_params; // параметры битрейта для p2p-соединения

    const struct mpipe_track_params_t** tracks;

} mpipe_device_params_rtc_t;

typedef void (*mpipe_cb_device_command_t)(const struct mpipe_media_device_t* device
                                         , const struct mpipe_command_t* command);

typedef void (*mpipe_cb_device_state_t)(const struct mpipe_media_device_t* device
                                        , enum mpipe_device_state_t device_state);

typedef void (*mpipe_cb_device_stream_t)(const struct mpipe_media_device_t* device
                                         , const struct mpipe_media_stream_t* stream
                                         , enum mpipe_stream_state_t stream_state);

typedef void (*mpipe_cb_device_frame_t)(const struct mpipe_media_device_t* device
                                         , const struct mpipe_media_stream_t* stream
                                         , const struct mpipe_media_frame_t* media_frame);

// блок описания функций обратного вызова
typedef struct mpipe_device_cb_t
{
    mpipe_cb_device_command_t           on_command;     // команды от устройства
    mpipe_cb_device_state_t             on_state;       // состояние устройства
    mpipe_cb_device_stream_t            on_stream;      // добавление/удаление/идменение стримов
    mpipe_cb_device_frame_t             on_frame;       // медиафрейм от стрима устройства

} mpipe_device_cb_t;

typedef struct mpipe_device_params_t
{
    enum mpipe_device_type_t            device_type;    // тип устройства

    union
    {
        mpipe_device_params_input_t     input;
        mpipe_device_params_output_t    output;
        mpipe_device_params_rtc_t       rtc;
    }                                   params;         // блок параментров зависящий от типа
    void*                               user_data;      // польовательский контекст

    mpipe_device_cb_t                   cb;

} mpipe_device_params_t;

// создать устройство и получить его дескриптор
EXPORT_API mpipe_handle_t mpipe_device_create(const struct mpipe_device_params_t* params);

// управление устройством (открыть/запустить/остановить/закрыть)
EXPORT_API enum mpipe_result_t mpipe_device_control(mpipe_handle_t handle
                                         , enum mpipe_device_control_id_t control_id);

// получить информацию о устройстве (считать описатель с текущим состоянием)
EXPORT_API enum mpipe_result_t mpipe_device_get_info(mpipe_handle_t handle
                                                    , struct mpipe_media_device_t* media_device);

// переслать команду устройству
EXPORT_API enum mpipe_result_t mpipe_device_send_command(mpipe_handle_t handle
                                                         , const struct mpipe_command_t* command);

EXTERN_C_END


#endif // MPIPE_API_DEVICE_H
