#ifndef MPIPE_OBJECT_SERIALIZE_DEFS_H
#define MPIPE_OBJECT_SERIALIZE_DEFS_H

#include "option/option_utils.h"
#include "option/option_writer.h"

namespace mpipe
{

namespace detail
{

template<typename T>
inline i_option::u_ptr_t serialize(const T& value)
{
    if (auto o = utils::option::create_option(option_type_t::object))
    {
        if (utils::option::serialize(*o, value))
        {
            return o;
        }
    }

    return nullptr;
}

template<typename T>
inline i_option::u_ptr_t serialize_array(const std::vector<T>& value)
{
    if (auto o = utils::option::create_option(option_type_t::array))
    {
        auto& array = static_cast<i_option_value<i_option::u_array_t>&>(*o).get();
        for (const auto& v : value)
        {
            if (auto p = serialize(v))
            {
                array.emplace_back(std::move(p));
            }
        }

        if (!array.empty())
        {
            return o;
        }
    }

    return nullptr;
}

template<typename T>
bool serialize_array(i_option& option, const std::vector<T>& values)
{
    if (option.type() == option_type_t::array)
    {
        auto& array = static_cast<i_option_value<i_option::u_array_t>&>(option).get();
        for (const auto& v : values)
        {
            if (auto p = serialize(v))
            {
                array.emplace_back(std::move(p));
            }
        }
    }
    return false;

}

template<typename T>
inline bool deserialize_array(const i_option& options, std::vector<T>& values)
{
    if (options.type() == option_type_t::array)
    {
        values.clear();
        auto& array = static_cast<const i_option_value<i_option::u_array_t>&>(options).get();
        for (const auto& o : array)
        {
            if (o)
            {
                T v = {};
                if (utils::option::deserialize(*o, v))
                {
                    values.emplace_back(std::move(v));
                }
            }
        }
    }

    return !values.empty();
}


}

#define __declare_object_serializer(object_type)\
    template<> option_type_t utils::option::get_value_type<object_type>() { return option_type_t::object; }\
    template<> i_option::u_ptr_t utils::option::serialize(const object_type& value) { return detail::serialize(value); }\
    template<> i_option::u_ptr_t utils::option::serialize(const std::vector<object_type>& value) { return detail::serialize_array(value); }\
    template<> bool utils::option::serialize(i_option& option, const std::vector<object_type>& values) { return detail::serialize_array(option, values);  }

#define __declare_object_serializer_begin(object_type)\
    __declare_object_serializer(object_type)\
    template<> bool utils::option::serialize(i_option& option, const object_type& value)\
    {\
        utils::option::option_writer writer(option);\
        bool ok = true;
#define __declare_object_serializer_end() return ok; }

#define __declare_serialize_object_member_name(member, name)\
    ok = ok && writer.set(name, value.member);

#define __declare_serialize_object_member(member)\
    __declare_serialize_object_member_name(member, #member)

#define __declare_object_deserializer_begin(object_type)\
    template<> bool utils::option::deserialize(const i_option& option, std::vector<object_type>& values) { return detail::deserialize_array(option, values); }\
    template<> bool utils::option::deserialize(const i_option& option, object_type& value)\
    {\
        utils::option::option_reader reader(option);\
        bool ok = false;
#define __declare_object_deserializer_end() return ok; }

#define __declare_deserialize_object_member_name(member, name)\
    ok |= reader.get(name, value.member);

#define __declare_deserialize_object_member(member)\
    __declare_deserialize_object_member_name(member, #member)

}

#endif // MPIPE_OBJECT_SERIALIZE_DEFS_H
